#include "stdafx.h"
#include "resource.h"
#include <time.h>
#include <ddraw.h>
#define D3D_OVERLOADS
#include <d3d.h>

const float M_2PI = 6.283185307179586476925286766559f;

#define NUM_DONUTS		4

#define R1				0.5f
#define R2				1.0f
#define NUM_SEG1		32
#define NUM_SEG2		32
#define NUM_VERTEX		6144

#define MAX_X			4.0f
#define MAX_Y			3.0f
#define MAX_Z			4.0f

#define MAX_SPEED		0.1f
#define MIN_SPEED		0.0125f

LPDIRECTDRAW lpDD = NULL;
LPDIRECTDRAWSURFACE	lpDDSZBuffer = NULL;
LPDIRECTDRAWSURFACE	lpDDSPrimary = NULL;
LPDIRECT3D2 lpD3D = NULL;
LPDIRECT3DDEVICE2 lpD3DDevice = NULL;
LPDIRECT3DVIEWPORT2 lpD3DViewport = NULL;
LPDIRECT3DMATERIAL2 lpD3DMaterial = NULL;
LPDIRECT3DLIGHT lpD3DLight = NULL;
D3DMATERIAL mtrl[NUM_DONUTS];

D3DVERTEX *Torus = NULL;

HWND hWnd;
CHAR szWindowClass[] = "FirstD3D";
CHAR szAppTitle[] = "My first Direct3D application";

struct DonutData {
	float angle;
	float x, y, z;
	float dx, dy, dz;
} donuts[NUM_DONUTS];

#define RES_X	800
#define RES_Y	600
#define BPP		32

//float fDensity = 0.66f;
//float fStart = -MAX_Z;
//float fEnd = MAX_Z;

void FreeObjects()
{
	if (lpDD) 
	{
		if (lpD3D)
		{
			lpD3D->Release();
			lpD3D = NULL;
		}
		if (lpDDSPrimary)
		{
			lpDDSPrimary->Release();
			lpDDSPrimary = NULL;
		}
		if (lpDDSZBuffer)
		{
			lpDDSZBuffer->Release();
			lpDDSZBuffer = NULL;
		}
		if (lpD3DDevice)
		{
			lpD3DDevice->Release();
			lpD3DDevice = NULL;
		}
		if (lpD3DViewport)
		{
			lpD3DViewport->Release();
			lpD3DViewport = NULL;
		}
		if (lpD3DMaterial)
		{
			lpD3DMaterial->Release();
			lpD3DMaterial = NULL;
		}
		if (lpD3DLight)
		{
			lpD3DLight->Release();
			lpD3DLight = NULL;
		}
		lpDD->Release();
		lpDD = NULL;
	}
	if (Torus) 
	{
		delete Torus;
		Torus = NULL;
	}
}

bool Failed()
{
	MessageBox(hWnd, "Direct3D init failed!", szAppTitle, MB_OK | MB_ICONSTOP);
	return false;
}

D3DVERTEX TorusVertex(int i, int j)
{
	D3DVERTEX v;
	float c1 = M_2PI * i / NUM_SEG1;
	float c2 = M_2PI * j / NUM_SEG2;
	float c = (float)cos(c1);
	float s = (float)sin(c1);
	float x, nx;

	nx = (float)cos(c2);
	x = nx * R1 + R2;
	v.nz = (float)sin(c2);
	v.z = v.nz * R1;

	v.nx = nx * c;
	v.x = x * c;
	v.ny = nx * s;
	v.y = x * s;

	return v;
}

void InitGeometry()
{
	int i, j, p = 0, m, n;

	Torus = new D3DVERTEX[NUM_VERTEX];

    for (i = 0; i < NUM_SEG1; i++)
	{
		m = i + 1;
		for (j = 0; j < NUM_SEG2; j++, p += 6)
		{
			n = j + 1;
			Torus[p] = TorusVertex(i, j);
			Torus[p + 2] = TorusVertex(i, n);
			Torus[p + 1] = TorusVertex(m, j);
			Torus[p + 3] = TorusVertex(i, n);
			Torus[p + 5] = TorusVertex(m, n);
			Torus[p + 4] = TorusVertex(m, j);
		}
	}
}

void InitDonuts()
{
	int i = 0;

	for (; i < NUM_DONUTS; i++)
	{
	    ZeroMemory(&mtrl[i], sizeof(D3DMATERIAL));
	    mtrl[i].dwSize = sizeof(D3DMATERIAL);
		donuts[i].angle = M_2PI * rand() / RAND_MAX;
		donuts[i].x = 2.0f * rand() / RAND_MAX * MAX_X - MAX_X;
		donuts[i].y = 2.0f * rand() / RAND_MAX * MAX_Y - MAX_Y;
		donuts[i].z = 2.0f * rand() / RAND_MAX * MAX_Z - MAX_Z;
		donuts[i].dx = 2.0f * MAX_SPEED * rand() / RAND_MAX - MAX_SPEED;
		donuts[i].dy = 2.0f * MAX_SPEED * rand() / RAND_MAX - MAX_SPEED;
		donuts[i].dz = 2.0f * MAX_SPEED * rand() / RAND_MAX - MAX_SPEED;
	}

    mtrl[0].dcvAmbient.r = 0.1f;
    mtrl[0].dcvAmbient.g = 0.0f;
    mtrl[0].dcvAmbient.b = 0.0f;
    mtrl[0].dcvDiffuse.r = 0.9f;
    mtrl[0].dcvDiffuse.g = 1.0f;
    mtrl[0].dcvDiffuse.b = 0.0f;

    mtrl[1].dcvAmbient.r = 0.0f;
    mtrl[1].dcvAmbient.g = 0.0f;
    mtrl[1].dcvAmbient.b = 0.1f;
    mtrl[1].dcvDiffuse.r = 1.0f;
    mtrl[1].dcvDiffuse.g = 0.0f;
    mtrl[1].dcvDiffuse.b = 0.0f;

    mtrl[2].dcvAmbient.r = 0.0f;
    mtrl[2].dcvAmbient.g = 0.1f;
    mtrl[2].dcvAmbient.b = 0.0f;
    mtrl[2].dcvDiffuse.r = 1.0f;
    mtrl[2].dcvDiffuse.g = 0.9f;
    mtrl[2].dcvDiffuse.b = 1.0f;

    mtrl[3].dcvAmbient.r = 0.1f;
    mtrl[3].dcvAmbient.g = 0.1f;
    mtrl[3].dcvAmbient.b = 0.0f;
    mtrl[3].dcvDiffuse.r = 0.0f;
    mtrl[3].dcvDiffuse.g = 0.0f;
    mtrl[3].dcvDiffuse.b = 1.0f;
}

bool SetUpMatrices()
{
	D3DMATRIX mat;

	mat._11 = mat._22 = mat._33 = mat._44 = 1.0f;
    mat._12 = mat._13 = mat._14 = mat._41 = 0.0f;
    mat._21 = mat._23 = mat._24 = mat._42 = 0.0f;
    mat._31 = mat._32 = mat._34 = mat._43 = 0.0f;
    
	D3DMATRIX matWorld = mat;
    if (lpD3DDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, &matWorld) != D3D_OK)
 		return Failed();

	D3DMATRIX matView = mat;
    matView._43 = 10.0f;
    if (lpD3DDevice->SetTransform(D3DTRANSFORMSTATE_VIEW, &matView) != D3D_OK)
 		return Failed();
 
	D3DMATRIX matProj = mat;
    matProj._11 =  2.0f;
    matProj._22 =  (float)RES_X / RES_Y * 2.0f;
    matProj._34 =  1.0f;
    matProj._43 = -1.0f;
    matProj._44 =  0.0f;

    if (lpD3DDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &matProj) != D3D_OK)
 		return Failed();

	return true;
}

bool InitDirect3D()
{
    DDSURFACEDESC ddsd;
	D3DVIEWPORT2 viewData;
	D3DLIGHT light;

	if (DirectDrawCreate(NULL, &lpDD, NULL) != DD_OK)
		return Failed();

	if (lpDD->QueryInterface(IID_IDirect3D2, (void **)&lpD3D) != DD_OK)
		return Failed();

    if (lpDD->SetCooperativeLevel(hWnd, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE) != DD_OK)
		return Failed();

	if (lpDD->SetDisplayMode(RES_X, RES_Y, BPP) != DD_OK)
 		return Failed();

    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.dwBackBufferCount = 1;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP | DDSCAPS_3DDEVICE | DDSCAPS_VIDEOMEMORY;

    if (lpDD->CreateSurface(&ddsd, &lpDDSPrimary, NULL) != DD_OK)
 		return Failed();

    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
    ddsd.dwWidth = RES_X;
    ddsd.dwHeight = RES_Y;
    ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | DDSCAPS_VIDEOMEMORY;
	ddsd.ddpfPixelFormat.dwSize = sizeof(ddsd.ddpfPixelFormat);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_ZBUFFER;
	ddsd.ddpfPixelFormat.dwZBufferBitDepth = 32;

    if (lpDD->CreateSurface(&ddsd, &lpDDSZBuffer, NULL) != DD_OK)
		return Failed();

	if (lpDDSPrimary->AddAttachedSurface(lpDDSZBuffer) != D3D_OK)
		return Failed();

	if (lpD3D->CreateDevice(IID_IDirect3DHALDevice, lpDDSPrimary, &lpD3DDevice) != D3D_OK)
 		return Failed();

	if (lpD3DDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, TRUE) != D3D_OK)
		return Failed();

	if (lpD3DDevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESSEQUAL) != D3D_OK)
		return Failed();

	if (lpD3DDevice->SetRenderState(D3DRENDERSTATE_SHADEMODE, D3DSHADE_GOURAUD) != D3D_OK)
		return Failed();

	if (lpD3D->CreateViewport(&lpD3DViewport, NULL) != D3D_OK)
 		return Failed();

	if (lpD3DDevice->AddViewport(lpD3DViewport) != D3D_OK)
 		return Failed();

	ZeroMemory(&viewData, sizeof(D3DVIEWPORT2));
    viewData.dwSize = sizeof(D3DVIEWPORT2);
    viewData.dwX = 0;
    viewData.dwY = 0;
    viewData.dwWidth  = RES_X;
    viewData.dwHeight = RES_Y;
    viewData.dvClipX  = -1.0f;
    viewData.dvClipY  =  1.0;
    viewData.dvClipWidth  = 2.0f;
    viewData.dvClipHeight = 2.0f;
    viewData.dvMinZ = 0.0f;
    viewData.dvMaxZ = 1.0f;
 
	if (lpD3DViewport->SetViewport2(&viewData) != D3D_OK)
 		return Failed();

    if (lpD3DDevice->SetCurrentViewport(lpD3DViewport) != D3D_OK)
 		return Failed();

	if (lpD3D->CreateMaterial(&lpD3DMaterial, NULL) != D3D_OK)
 		return Failed();

	if (lpD3D->CreateLight(&lpD3DLight, NULL) != D3D_OK)
 		return Failed();

	ZeroMemory(&light, sizeof(D3DLIGHT));
	light.dwSize = sizeof(D3DLIGHT);
	light.dltType = D3DLIGHT_POINT;
	light.dvPosition = D3DVECTOR(4.0f, 4.0f, -4.0f);
	light.dcvColor.r = 1.0f;
	light.dcvColor.g = 1.0f;
	light.dcvColor.b = 1.0f;
	light.dvRange = D3DLIGHT_RANGE_MAX;

	if (lpD3DLight->SetLight(&light) != D3D_OK)
 		return Failed();

	if (lpD3DViewport->AddLight(lpD3DLight) != D3D_OK)
 		return Failed();

	if (lpD3DDevice->SetLightState(D3DLIGHTSTATE_AMBIENT, 0xFFFFFFFF) != D3D_OK)
 		return Failed();

	if (!SetUpMatrices()) return false;

	InitGeometry();
	InitDonuts();

	return true;
}

float SpeedChange()
{
	return 0.4f * rand() / RAND_MAX + 0.8f;
}

void Range(float min, float max, float &val)
{
	if (val > MAX_SPEED) val = MAX_SPEED;
	if (val < -MAX_SPEED) val = -MAX_SPEED;
	if (val < MIN_SPEED && val > 0) val = MIN_SPEED;
	if (val > -MIN_SPEED && val < 0) val = -MIN_SPEED;
}

void UpdateFrame(int index)
{
	D3DMATRIX matSpin; 
	DonutData &d = donuts[index];
	float c = (float)cos(d.angle);
	float s = (float)sin(d.angle);

	matSpin._11 = c * c;
	matSpin._12 = c * s;
	matSpin._13 = s;
	matSpin._14 = 0.0f;
	matSpin._21 = -s;
	matSpin._22 = c;
	matSpin._23 = 0.0f;
	matSpin._24 = 0.0f;
	matSpin._31 = -c * s;
	matSpin._32 = -s * s;
	matSpin._33 = c;
	matSpin._34 = 0.0f;
	matSpin._41 = d.x;
	matSpin._42 = d.y;
	matSpin._43 = d.z;
	matSpin._44 = 1.0f;
 
    lpD3DDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, &matSpin);

	d.x += d.dx;
	d.y += d.dy;
	d.z += d.dz;

	if (d.x >= MAX_X || d.x <= -MAX_X) { d.x -= d.dx; d.dx = -d.dx * SpeedChange(); }
	if (d.y >= MAX_Y || d.y <= -MAX_Y) { d.y -= d.dy; d.dy = -d.dy * SpeedChange(); }
	if (d.z >= MAX_Z || d.z <= -MAX_Z) { d.z -= d.dz; d.dz = -d.dz * SpeedChange(); }

	Range(MIN_SPEED, MAX_SPEED, d.dx);
	Range(MIN_SPEED, MAX_SPEED, d.dy);
	Range(MIN_SPEED, MAX_SPEED, d.dz);

	d.angle += 0.05f;
	if (d.angle > M_2PI) d.angle -= M_2PI;
}

void SetMaterial(D3DMATERIAL mtrl)
{
	D3DMATERIALHANDLE hmtrl;

    lpD3DMaterial->SetMaterial(&mtrl);
	lpD3DMaterial->GetHandle(lpD3DDevice, &hmtrl);
    lpD3DDevice->SetLightState(D3DLIGHTSTATE_MATERIAL, hmtrl);
}

void DrawFrame()
{
	D3DRECT Rect;

	Rect.x1 = 0;
	Rect.y1 = 0;
	Rect.x2 = RES_X;
	Rect.y2 = RES_Y;

/*	lpD3DDevice->SetRenderState(D3DRENDERSTATE_TEXTUREPERSPECTIVE, TRUE);
	lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, TRUE);
	lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGCOLOR, 0xFFFFFFFF);
	lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_LINEAR);
	lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGTABLESTART, *(DWORD *)(&fStart));
	lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGTABLEEND, *(DWORD *)(&fEnd));*/

//	if (lpD3DDevice->SetRenderState(D3DRENDERSTATE_FOGTABLEDENSITY, *(DWORD *)(&fDensity)) != D3D_OK)
//		return Failed();

	if (lpD3DViewport->Clear(1, &Rect, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER) != D3D_OK)
		return;
 
	if (lpD3DDevice->BeginScene() != D3D_OK)
		return;

	for (int i = 0; i < NUM_DONUTS; i++)
	{
		UpdateFrame(i);
		SetMaterial(mtrl[i]);
		lpD3DDevice->DrawPrimitive(D3DPT_TRIANGLELIST, D3DVT_VERTEX, Torus, NUM_VERTEX, NULL);
	}

	lpD3DDevice->EndScene();

	if (!lpDDSPrimary) return;

	lpDDSPrimary->Flip(NULL, DDFLIP_WAIT);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) 
	{
		case WM_DESTROY:
			FreeObjects();
			PostQuitMessage(0);
			break;
		case WM_KEYDOWN:
			if ((int)wParam == VK_ESCAPE)
				PostMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	WNDCLASSEX wcex;

	srand((unsigned)time(NULL));

	wcex.cbSize			= sizeof(WNDCLASSEX); 
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_FIRSTDIRECT3D);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	RegisterClassEx(&wcex);

	hWnd = CreateWindow(szWindowClass, szAppTitle, WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if (!InitDirect3D()) return 0;

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	while (1)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
		    if(!GetMessage(&msg, NULL, 0, 0)) return msg.wParam;
		    TranslateMessage(&msg); 
		    DispatchMessage(&msg);
		}
        else DrawFrame();
	}

	return 0;
}
